--[[
    =========
    = MIMIC =  yes is tiny box
    =========
    
    Inspired by the character Emerl, and to some extent, stat randomizers, I give you... Mimic. What does it do? It makes you
    mimic characters, of course! At the start of the race/battle, a random character will be chosen from the pool of characters
    you have added. You will mimic this character for the duration of the race/battle (by default), both stats and voices.
    However, you can make things a bit crazier than this.
    
    
    
    There are 4 game modes included, all of which are fully compatible with one another. So, you can mix/match the various modes
    to create up to ~~wait did i do my math right~~ 15 different combinations. Along with adjustable intervals and cooldowns, you
    have a lot of room for customization.
    
    
    Respawn mode!
        This causes you to change what character you're mimicing on respawn. The character you get is random. It has a cooldown
        of 10 seconds by default, so don't think you can abuse death until you get a character you like. The cooldown is
        adjustable, however, so don't fret if you dislike this. (and yes... you can set it to 0 if you'd like to be lame)
        
        Related console vars:
        
            mimic_respawn: On/Off             (Default: Off)
            mimic_respawncooldown: Unsigned   (Default: 10 )
    
    Lap mode!
        This causes you to change what character you're mimicing every lap. The character you get is random.
        
        Related console vars:
        
            mimic_lap: On/Off                 (Default: Off)
    
    Manual mode!
        This adds the ability for you to change mimics at will. This is done by tapping Custom 3. The character you get is random, 
        and this has a cooldown of 25 seconds by default. Try to be wise with your rolls and perhaps you'll get what you need at
        just the right time. The cooldown for this mode is also adjustable.
        
        Related console vars:
        
            mimic_manual: On/Off              (Default: Off)
            mimic_manualcooldown: Unsigned    (Default: 25 )
    
    Random mode!
        Perhaps the most exciting mode. This causes you to randomly change mimics throughout the race/battle. By default, this
        happens on a random interval between 15-30 seconds (decided each time you change mimics). You can adjust this interval
        as well, and if you'd like to make the change time static, simply make the min and max the same.
        
        Related console vars:
        
            mimic_random: On/Off              (Default: Off)
            mimic_randommin: Unsigned         (Default: 15 )
            mimic_randommax: Unsigned         (Default: 30 )
    
    
    
    Additionally, for those that might dislike the fact that you mimic the sounds of said mimicked character, or dislike having
    sounds in general (and possibly HUDs), you've been accounted for.
    
        This allows you to disable mimicking sounds. Note that turning this off will make it harder for you to know who the other
        players really are!
            
            mimic_sounds: On/Off                  (Default: On )
        
        This lets you disable indicator sound effects, for those of you that might find it a tad annoying. You'll have to pay
        better attention to your HUD without this, however.
            
            mimic_indicatorsounds: On/Off         (Default: On )
        
        This allows you to disable the mimic HUD. Why would you ever want to do this?
            
            mimic_hud: On/Off                     (Default: On )
    
    
    
    Also, Splitscreen is fully supported! Bless you toaster for exposing cameras
    
    
    
    Full Console commands/variable documentation:
    
        Variables:    
            Net:
                mimic                : On/Off               Allows you to toggle mimic on or off
                    
                mimic_respawn        : On/Off               Allows you to toggle respawn mode on or off
                mimic_lap            : On/Off               Allows you to toggle lap mode on or off
                mimic_manual         : On/Off               Allows you to toggle manual mode on or off
                mimic_random         : On/Off               Allows you to toggle random mode on or off
             
                mimic_respawncooldown: Unsigned             Allows you to adjust the respawn mode cooldown
                mimic_manualcooldown : Unsigned             Allows you to adjust the manual mode cooldown
                mimic_randommin      : Unsigned             Allows you to adjust the random mode minimum interval
                mimic_randommax      : Unsigned             Allows you to adjust the random mode maximum interval
        
            Local:
                mimic_hud            : On/Off               Allows you to toggle the mimic HUD on or off
                mimic_sounds         : On/Off               Allows you to toggle mimicing character sounds on or off
                mimic_sounds         : On/Off               Allows you to toggle indicator sound effects on or off
                
        Commands:
            Local:
                mimic_help           :                      Provides you with useful help information while in-game




    Coded by Namolos (aka: Nathan). Feel free to use this code as a reference for yourself. And if you borrow anything
    from this project (even though I don't recommend doing that lmao), please give me credit!
--]]    

--Help command
local function gethelp(p) --run
    CONS_Printf(p, "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@&(////////////////////////(&@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"..
                   "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&&(/////////////////////////%@@@@@@@@@@@@@/////////////////\n"..
                   "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@#//////////////////////////&@@@@@@@@@@@#////////////////////\n"..
                   "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&&(//////////////////////////&@@@@@@@@@%////////////////////////\n"..
                   "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%//////////////////////////(&@@@@@@@@%///////////////////////////\n"..
                   "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%//////////////////////////(@@@@@@@@##%(*,,,,,*(%(/////////////////\n"..
                   "&@@@@@@@@@@@@&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%//////////////////////////(&@@@@@@&%*,,,,,,,,,,,,,,,(#//////////////\n"..
                   "@@@@@@@@@@@@@@@@@@@&%%%%%%%%%%%%%%%%%%%%%%%%%%%%&(/////////////////////////(&@@@@@/,,,,,,,,,,,,,,,,,,,,,%//////////////\n"..
                   "@@@(///////////(#&@@@&%%%%%%%%%%%%%%%%%%%%%%%#/////////////////////////(&@@@@@#((,,,,,,,,,,,,,,,,,,,,,,,,*%////////////\n"..
                   "//((####(//////////////(%@&%%%%%%%%%%%%%%%%%%&//////////////////////////&@@@@%//(/,,,,,,,,,,,,,,,,,,,,,,,,,,,((////////\n"..
                   "@@@@@@@@@@@@@@#(///////////(@%%%%%%%%%%%%%%%&/////////////////////////%@@@@(///(*,,,,,,,,,,,,,,,,,,,,,,,,,,,,,/#///////\n"..
                   "@@@@@@@@@@@@@@@@@@%///////////#&%%%%%%%%%%%#////////////////////////%@@@&//////(*,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,/(//////\n"..
                   "@@@@@&&&&&&@@@@@@@@@@@(/////////(&%%%%%%%%#///////////////////////&@@&(///////%,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,((/////\n"..
                   "////////////(((//#&@@@@@@(////////(&%%%%&(///////////////(#//%%%%%#///////////#,,,,,,,,,,(&%%%%%%%*,,,,,,,,,,,,,,%/////\n"..
                   "////////#/,,,,,,,,*#%/%@@@@&////////(&%%(////////////////////###/////////////(*,,,,,,,,*&%%%%%%%%%%&*,,,,,,,,,,,,*#////\n"..
                   "/////((*,,,,,,,,,,,,,,/#/(&@@@(///////%#/////////////////////////////////////%,,,,,,,,*%%%%%%%%%%%%%&*,,,,,,,,,,,*#////\n"..
                   "////%*,,,,,,,,,,,,,,,,,,*#//(&@@(////////////////////////////////////////////%,,,,,,,,#%%%%%%%%%%%%%%&,,,,,,,,,,,,*#///\n"..
                   "///%,,,,,,,,,,,,,,,,,,,,,,#/////&&///////////////////////////////////////////%*,,,,,,,&%%%%%%%%%%%%%%%%,,,,,,,,,,,,%///\n"..
                   "//%*,,,,,,,,,,,,,,,,,,,,,,**#/////#%/////////////////////////////////////////%,,,,,,,,&%%%%%&&%%%%%%%%&,,,,,,,,,,,*%///\n"..
                   "/(/,,,,,,,,,,,,,,,,,,,,,,,,,*#///////////////////////////////////////////////%,,,,,,,,&%%%%%@@%%%%%%%%&*,,,,,,,,,,,#///\n"..
                   "/%,,,,,,,,,,,,(%&%#,,,,,,,,,,*(//////((//////////////////////////////////////#,,,,,,,,%%%%%%%&%%%%%%%%&*,,,,,,,,,,,#///\n"..
                   "(/*,,,,,,,,,%%%%%%%%%(,,,,,,,,(////////%/////////////////////////////////////(*,,,,,,,*%%%%%%%%%%%%%%%&,,,,,,,,,,,,%///\n"..
                   "#**,,,,,,,,&%%%%%%%%%%&*,,,,,,,%////////%/////////////////////////////////////%,,,,,,,,#%%%%%%%%%%%%%%*,,,,,,,,,,,*%///\n"..
                   "%*,,,,,,,,%%%%%%%%%%%%%&*,,,,,,*(////////%////////////////////////////////////(*,,,,,,,,#%%%%%%%%%%%%(,,,,,,,,,,,,/(///\n"..
                   "%*,,,,,,,*%%%%%%%%%%%%%%%,,,,,,*%/////////%////////////////////////////////////&,,,,,,,,,*&%%%%%%%%&*,,,,,,,,,,,,,#////\n"..
                   "%*,,,,,,,(%%%%%%@@%%%%%%&,,,,,,,/(////////((////////////////////////////////////%,,,,,,,,,,,*%&&%/,,,,,,,,,,,,,,,((////\n"..
                   "(*,,,,,,,(%%%%%%@@&%%%%%%/,,,,,,*#/////////%/////////////////////////////////////%,,,,,,,,,,,,,,,,,,**/(#%%%%%%%%%#(///\n"..
                   "/#*,,,,,,*%%%%%%&@%%%%%%%(,,,,,,*#/////////%//////////////////////////////////////%,,,,,,,,,/(%#(//////////////////////\n"..
                   "/%**,,,,,*%%%%%%%%%%%%%%%/,,,,,,*#/////////%///////////////////////////////////////((,/%(//////////////////////////////\n"..
                   "//(*,,,,,,*%%%%%%%%%%%%%&,,,,,,,/(/////////(///////////////////////////////////////%(//////////////////////////////////\n"..
                   "//(*,,,,,,,*%%%%%%%%%%%&,,,,,,,*#/////////%////////////////////////////////////////////////////////////////////////////\n"..
                   "///%*,,,,,,,,*%%%%%%%&(,*,,,,,,*#////////(/////////////////////////////////////////////////////////////////////////////\n"..
                   "////%*,,,,,,,,,,,***,,,,,,,,,**%////////((/////////////////////////////////////////////////////////////////////////////\n"..
                   "/////#/,,,,,,,,,*/#%%########%%%#(/////%///////////////////////////////////////////////////////////////////////////////\n"..
                   "///////#(#%(///////////////////////(%/////////////////////////(%///////////////////////////////////////////////////////\n"..
                   "/////##///////////////////////////#//////////////////////////////#/////////////////////////////////////////////////////\n"..
                   "//////////////////////////////////#//////////////////////////////%/////////////////////////////////////////////////////\n"..
                   "//////////////////////////////////#//////////////////////////////&/////////////////////////////////////////////////////\n"..
                   "//////////////////////////////////%/////////////////////////////&(/////////////////////////////////////////////////////\n"..
                   "///////////////////////////////////#////////////////#(((#(/////////////////////////////////////////////////////////////\n"..
                   "////////////////////////////////////(//////////////////////////////////////////////////////////////////////////////////\n"..
                   "///////////////////////////////////////((##%%#/////////////////////////////////////////////////////////////////////////\n"..
                   "////////////////////////////////////////////////%(/////////////////////////////////////////////////////////////////////\n"..
                   "///////////////////////////////////////////////////////////////////////////////////(#//////////////////////////////////"
                   )

    --This block is only useful for admins
    if (IsPlayerAdmin(p) or (p == server))
        CONS_Printf(p, "\nYou have the following net vars at your disposal:\n\n"..
                       "    On/Off Toggles                                                        Adjustable Values\n"..
                       "----------------------                                              -----------------------------\n"..
                       "    mimic                Enable/Disable mimic                           mimic_respawncooldown        Allows you to set the respawn cooldown (in seconds)\n"..
                       "    mimic_respawn        Enable/Disable respawn mode                    mimic_manualcooldown         Allows you to set the manual cooldown (in seconds)\n"..
                       "    mimic_lap            Enable/Disable lap mode                        mimic_randommin              Allows you to set the minumum possible interval for random mode (in seconds)\n"..
                       "    mimic_manual         Enable/Disable manual mode                     mimic_randommax              Allows you to set the maximum possible interval for random mode (in seconds)\n"..
                       "    mimic_random         Enable/Disable random mode"
                   )               
    end
    
    --This info is useful for everyone
    CONS_Printf(p, "\nYou have the following local vars you can adjust:\n\n"..
                   "    On/Off Toggles\n"..
                   "----------------------\n"..
                   "    mimic_hud               Enable/Disable mimic HUD\n"..
                   "    mimic_sounds            Enable/Disable sound effect mimicing\n"..
                   "    mimic_indicatorsounds   Enable/Disable indicator sound effects\n"          
               )
end

COM_AddCommand("mimic_help", gethelp)


--constants
--time constants
local INDICATORTIME = 3*TICRATE            --Initial indicator time
local MANUALSTARTINGDELAY = 7*TICRATE      --Initial manual mode delay time
local MANUALINSTRUCTIONTIME = 16*TICRATE   --Delay time for manual mode instruction

--sound constants
local INDICATORSOUND = sfx_kc5e            --Indicator sound effect
local MANUALERRORSOUND = sfx_kc43          --Indicator error sound effect

--hud constants
local HUDY = 170        --oldvals: 60, 177
local HUDFLAGS = V_30TRANS|V_SNAPTOLEFT|V_SNAPTOBOTTOM
local function HUDX()
	if acrobatics and ac_active then
		return 120
	else
		return 92
	end
end

--variables
local respawnCooldown = 10                 --The respawn cooldown, duh
local manualCooldown = 25                  --The manual cooldown, dummy
local randomMin = 15                       --The minimum possible interval for random mimic changes
local randomMax = 30                       --The maximum possible interval for random mimic changes
local knowstheirshit = false               --i am a very clever namer gib ribbon plz


--Sync these vars, just in case
addHook("NetVars", function(net)
    respawnCooldown = net(respawnCooldown)
    manualCooldown = net(manualCooldown)
    randomMin = net(randomMin)
    randomMax = net(randomMax)
end)


--CVar functions
--Sets the minimum possible interval for random mode in seconds (Default: 15 seconds)
local function setMin(m)
    randomMin = m.value
    
    if (randomMin > randomMax)
        CONS_Printf(server, "Wait, you made the min larger than the max (max: "..randomMax.."). I mean, this will work, but why do?")
    end
end

--Sets the maximum possible interval for random mode in seconds (Default: 30 seconds)
local function setMax(m)
    randomMax = m.value
    
    if (randomMax < randomMin)
        CONS_Printf(server, "Hold up, you made the max smaller than the min (min: "..randomMin.."). I mean, this will work, but why do?")
    end
end

--In the event that random mode is toggled midrace, this makes sure things get setup correctly
local function setRandomCountdown(c)
    if (c.value == 1)
        for p in players.iterate
            if (leveltime >= p.mimic.nextChange)
                p.mimic.hasChanged = true
            end
        end
    end
end

--Sets the cooldown time for respawn mode in seconds (Default: 10 seconds)
local function setRespawnCooldown(c)
    respawnCooldown = c.value
end

--Sets the cooldown time for manual mode in seconds (Default: 25 seconds)
local function setManualCooldown(c)
    manualCooldown = c.value
end

--Resets stats back to normal when Mimic is toggled off
local function resetStats(m)
    if (m.value == 0)
        for p in players.iterate
            p.kartspeed = skins[p.mo.skin].kartspeed
            p.kartweight = skins[p.mo.skin].kartweight
        end
    end
end


--CVars
--net (main switch)
local cv_mimicenabled = CV_RegisterVar({"mimic", "On", CV_CALL|CV_NETVAR|CV_SHOWMODIF, CV_OnOff, resetStats})
local cv_acromimic = CV_RegisterVar({"mimic_acrobatics", "On", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})

--net (modes)
local cv_respawnmimic = CV_RegisterVar({"mimic_respawn", "Off", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local cv_lapmimic = CV_RegisterVar({"mimic_laps", "On", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local cv_manualmimic = CV_RegisterVar({"mimic_manual", "On", CV_NETVAR|CV_SHOWMODIF, CV_OnOff})
local cv_randommimic = CV_RegisterVar({"mimic_random", "Off", CV_CALL|CV_NETVAR|CV_SHOWMODIF, CV_OnOff, setRandomCountdown})

--net (misc)
local cv_respawnmimiccooldown = CV_RegisterVar({"mimic_respawncooldown", "10", CV_CALL|CV_NETVAR|CV_SHOWMODIF, CV_Unsigned, setRespawnCooldown})
local cv_manualmimiccooldown = CV_RegisterVar({"mimic_manualcooldown", "25", CV_CALL|CV_NETVAR|CV_SHOWMODIF, CV_Unsigned, setManualCooldown})
local cv_randommimicmin = CV_RegisterVar({"mimic_randommin", "15", CV_CALL|CV_NETVAR|CV_SHOWMODIF, CV_Unsigned, setMin})
local cv_randommimicmax = CV_RegisterVar({"mimic_randommax", "30", CV_CALL|CV_NETVAR|CV_SHOWMODIF, CV_Unsigned, setMax})

--local
local cv_mimicstathud = CV_RegisterVar({"mimic_hud", "On", CV_SHOWMODIF, CV_OnOff})
local cv_mimicsounds = CV_RegisterVar({"mimic_sounds", "On", CV_SHOWMODIF, CV_OnOff})
local cv_indicatorsounds = CV_RegisterVar({"mimic_indicatorsounds", "On", CV_SHOWMODIF, CV_OnOff})

--main functions
--It... it picks a random skin, okay. As well as storing the last skin you had for possible rerolls
local function randomSkin(p)
    p.mimic.lastSkin = p.mimic.randomSkin
    p.mimic.randomSkin = A_RandomKey(#skins)
    
    --A single reroll in the event that you roll a duplicate... to give the illusion that it's more random than it is :x
    if (p.mimic.randomSkin == p.mimic.lastSkin)
       p.mimic.randomSkin = A_RandomKey(#skins)
    end
	
    p.mimic.lastAcro = p.mimic.randomAcro
    p.mimic.randomAcro = A_RandomRange(0, 3)
  
    if (p.mimic.randomAcro == p.mimic.lastAcro)
       p.mimic.randomAcro = A_RandomRange(0, 3)
    end
end

--This (hacky af) function allows you to replace one sound with another. Used here for sound mimicry
local function replaceSoundByMobj(m, s, r)
    if (S_SoundPlaying(m, s))
        S_StopSoundByID(m, s)
        S_StartSound(m, r)
    end
end

--This function handles stat mimic, featuring special conditions to avoid people cheesing with OoR stat characters
local function statMimic(p)
    --speed
    if (skins[p.mimic.randomSkin].kartspeed < 1)
        p.kartspeed = 1
    elseif (skins[p.mimic.randomSkin].kartspeed > 9)
        p.kartspeed = 9
    else
        p.kartspeed = skins[p.mimic.randomSkin].kartspeed
    end
    
    
    if (skins[p.mimic.randomSkin].kartweight < 1)
        p.kartweight = 1
    elseif (skins[p.mimic.randomSkin].kartweight > 9)
        p.kartweight = 9
    else
        p.kartweight = skins[p.mimic.randomSkin].kartweight
    end
end

local function acroMimic(p)
	if (acrobatics == nil) or (not cv_acromimic.value) then return end
	
    --CGB stuff
    if p.randomacrobatics then
		p.randomacrobatics = 0
		S_StartSound(nil, sfx_cdfm20, p)
		CONS_Printf(p, "\130MIMIC DETECTED! RANDOM Acrobatics Style deactivated...")
	end

    p.trickstyle = p.mimic.randomAcro
end

--The actual sound mimicry. Effectively replaces all of your skin sound effects with that of the character you're mimicing (albeit with a few oddities)
local function soundMimic(p)
    --KWIN     Oh. Well that's spectacular. RIP
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKWIN], skins[p.mimic.randomSkin].soundsid[SKSKWIN])
    
    --KLOSE    RIP to you too buddy
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKLOSE], skins[p.mimic.randomSkin].soundsid[SKSKLOSE])

    --KHURT1
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKPAN1], skins[p.mimic.randomSkin].soundsid[SKSKPAN1])

    --KHURT2
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKPAN2], skins[p.mimic.randomSkin].soundsid[SKSKPAN2])

    --KATTK1
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKATK1], skins[p.mimic.randomSkin].soundsid[SKSKATK1])

    --KATTK2
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKATK2], skins[p.mimic.randomSkin].soundsid[SKSKATK2])

    --KBOST1
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKBST1], skins[p.mimic.randomSkin].soundsid[SKSKBST1])

    --KBOST2
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKBST2], skins[p.mimic.randomSkin].soundsid[SKSKBST2])

    --KSLOW
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKSLOW], skins[p.mimic.randomSkin].soundsid[SKSKSLOW])

    --KHITEM
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKHITM], skins[p.mimic.randomSkin].soundsid[SKSKHITM])

    --KGLOAT
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKPOWR], skins[p.mimic.randomSkin].soundsid[SKSKPOWR])
end

--This entire "snowflake" block exists purely because your skins sounds are mapped over Sonic's, meaning trying to mimic Sonic as anyone BUT Sonic results in an infinite start/stop sound loop. Allow me to spare your ears

--God fucking damn it, Sonic. Why do you have to ruin shit?
freeslot("sfx_sdupl1",
         "sfx_sdupl2",
         "sfx_sdupl3",
         "sfx_sdupl4",
         "sfx_sdupl5",
         "sfx_sdupl6",
         "sfx_sdupl7",
         "sfx_sdupl8",
         "sfx_sdupl9",
         "sfx_sdupla",
         "sfx_sduplb")

--Disgusting snowflake workaround table with corresponding names for easier reading
local specialSnowflakeSonicSounds = {
    SKSKWIN  = sfx_sdupl1,  --864
    SKSKLOSE = sfx_sdupl2,  --865
    SKSKPAN1 = sfx_sdupl3,  --866
    SKSKPAN2 = sfx_sdupl4,  --867
    SKSKATK1 = sfx_sdupl5,  --868
    SKSKATK2 = sfx_sdupl6,  --869
    SKSKBST1 = sfx_sdupl7,  --870
    SKSKBST2 = sfx_sdupl8,  --871
    SKSKSLOW = sfx_sdupl9,  --872
    SKSKHITM = sfx_sdupla,  --873
    SKSKPOWR = sfx_sduplb   --874
}

--I hate everything about this functions existence
local function specialSnowflakeSonicMimic(p)
    --KWIN     Oh. Well that's spectacular. RIP
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKWIN], specialSnowflakeSonicSounds.SKSKWIN)
    
    --KLOSE    RIP to you too buddy
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKLOSE], specialSnowflakeSonicSounds.SKSKLOSE)

    --KHURT1
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKPAN1], specialSnowflakeSonicSounds.SKSKPAN1)

    --KHURT2
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKPAN2], specialSnowflakeSonicSounds.SKSKPAN2)

    --KATTK1
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKATK1], specialSnowflakeSonicSounds.SKSKATK1)

    --KATTK2
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKATK2], specialSnowflakeSonicSounds.SKSKATK2)

    --KBOST1
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKBST1], specialSnowflakeSonicSounds.SKSKBST1)

    --KBOST2
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKBST2], specialSnowflakeSonicSounds.SKSKBST2)

    --KSLOW
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKSLOW], specialSnowflakeSonicSounds.SKSKSLOW)

    --KHITEM
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKHITM], specialSnowflakeSonicSounds.SKSKHITM)

    --KGLOAT
    replaceSoundByMobj(p.mo, skins[p.mo.skin].soundsid[SKSKPOWR], specialSnowflakeSonicSounds.SKSKPOWR)             
end

--Indication sound to help the player know what character they're mimicing
local function indicatorSound(p)
    if (leveltime == p.mimic.newIndicator)
        S_StartSound(p.mo, INDICATORSOUND, p)
    end
    
    if (leveltime == p.mimic.newIndicator+(TICRATE/2))
        if (p.mimic.randomSkin == 0)  --snowflake check. Still hate this :)
            S_StartSound(p.mo, specialSnowflakeSonicSounds.SKSKSLOW, p)
        elseif (p.mo.skin == skins[p.mimic.randomSkin].name)  --Special case, matching skin
            S_StartSound(p.mo, sfx_kslow, p)
        else
            S_StartSound(p.mo, skins[p.mimic.randomSkin].soundsid[SKSKSLOW], p)
        end
    end
end


--for the safe or risky, I give you, respawn mimics mode!
local function respawnMimic(p)
    if (p.playerstate == PST_REBORN)
        if (leveltime > p.mimic.respawnDelay)
            p.mimic.hasChanged = true
            p.mimic.newIndicator = leveltime+1
            randomSkin(p)
            
            p.mimic.respawnDelay = leveltime+(respawnCooldown*TICRATE)
        end
    end
end

--for those that like to know what to expect, lap mimics mode!
local function lapMimic(p)
    if (p.mimic.lastLap < p.laps)
        p.mimic.hasChanged = true
        p.mimic.newIndicator = leveltime+1
        randomSkin(p)
        
        p.mimic.lastLap = p.laps
    end
end

--because why not? Manual mimics mode!
local function manualMimic(p)
    if (p.cmd.buttons & BT_CUSTOM3)
        if (leveltime > p.mimic.manualDelay)
            p.mimic.hasChanged = true
            p.mimic.newIndicator = leveltime+1
            randomSkin(p)

            p.mimic.manualDelay = leveltime+(manualCooldown*TICRATE)
            
            p.mimic.manualHudError = leveltime+TICRATE
            S_StopSoundByID(p.mo, MANUALERRORSOUND)
            
            --If the player does this in a timely manner, mark them as a knowledgeable reader and don't display the help text again :)
            if (leveltime < MANUALINSTRUCTIONTIME+(10*TICRATE))
                knowstheirshit = true
            end
            
        elseif ((leveltime > p.mimic.manualHudError) and not (S_SoundPlaying(p.mo, MANUALERRORSOUND)))
            if (cv_indicatorsounds.value)   --shhhhh, I'm lazy
                S_StartSound(p.mo, MANUALERRORSOUND, p)
            end
            p.mimic.manualHudError = leveltime+TICRATE
        end
    end
end

--for a truly bizarre experience, have random mimics mode!
local function randomMimic(p)
    if (p.mimic.hasChanged)  --Whenever this value gets set, go ahead and pick a new time to change mimics again
        p.mimic.nextChange = leveltime+(TICRATE*(A_RandomRange(randomMin, randomMax)))
        p.mimic.hasChanged = false
    end
                    
    if (leveltime == p.mimic.nextChange)
        p.mimic.hasChanged = true
        p.mimic.newIndicator = leveltime+1
        randomSkin(p)
    end
end

--What to do when the map loads. Duh?
addHook("MapLoad", do
    --Pick a random skin from the available skins to mimic, and initialize variables
    for p in players.iterate
        p.mimic = {           --table of mimic shit. Clear out with each map load/player join
        
		    lastSkin = 0,                            --The last skin this player mimiced. Used for rerolling upon getting a duplicate
		    randomSkin = A_RandomKey(#skins),        --The skin currently being mimiced. A random one is picked here anyway just so lastSkin gets set correctly in the upcoming randomSkin() call
		    lastAcro = 0,                            --The last skin this player mimiced. Used for rerolling upon getting a duplicate
		    randomAcro = A_RandomRange(0, 2),        --The skin currently being mimiced. A random one is picked here anyway just so lastSkin gets set correctly in the upcoming randomSkin() call
		    nextChange = 0,                          --For random mode. Whenever the leveltime matches this value, a mimic change happens
		    hasChanged = true,                       --Also for random mode. Whenever this value becomes true, the nextChange value gets set. This value gets set to true in all modes to ensure compatibility with random mode
		    newIndicator = INDICATORTIME,            --Used to control when indicator sounds play, as well as the flicker effect for a certain HUD graphic
		    respawnDelay = 0,                        --This value is used to check if the player has recently changed mimics on respawn. If the leveltime isn't larger than this value, they don't change
		    lastLap = p.laps,                        --This is used for lap mode. Simply put, whenever this value is smaller than the current lap, you change mimics and this increments
		    manualDelay = MANUALSTARTINGDELAY,       --Used for manual mode. Acts as a cooldown for manually changing mimics, as well as helps manipulate the corresponding HUD bits
		    manualHudError = 0                       --Also used for manual mode. Purely a tracker value for playing error sounds and giving a HUD effect to indicate said error
		}
        
        --Pick a random skin to start with
        randomSkin(p)
    end
end)


--Main thinkframe, handles stats/sounds as well as the various modes
addHook("ThinkFrame", do
    if (cv_mimicenabled.value)         --Ask not for whom the mode runs, it runs for thee
        if (leveltime > TICRATE)       --I know this looks pointless... but it fixes a very odd bug
	        for p in players.iterate do
	            if (p.mimic == nil)
	                p.mimic = {           --table of mimic shit. This is here for players joining late to get them setup
		            
		                lastSkin = 0,
		                randomSkin = A_RandomKey(#skins),
						lastAcro = 0,
						randomAcro = A_RandomRange(0, 2),
		                nextChange = 0,
		                hasChanged = true,
		                newIndicator = leveltime+INDICATORTIME,
		                respawnDelay = 0,
		                lastLap = p.laps,
		                manualDelay = MANUALSTARTINGDELAY,
		                manualHudError = 0
		            }

                    --Pick a random skin to start with
                    randomSkin(p)
	            end
	            if not (p.spectator)
                    --Stat mimicry
                    statMimic(p)
					acroMimic(p)

                    --Indicator sounds, to help players know what character they are mimicing
                    if (cv_indicatorsounds.value)
                        indicatorSound(p)
                    end

                    --The rest of the sounds, unless you're matching. Why remap if we're matching?
                    if (cv_mimicsounds.value)
                        if (p.mo.skin != skins[p.mimic.randomSkin].name)
                            --Special Sonic case, because for some reason he's a fucking special snowflake and breaks everything
                            if (p.mimic.randomSkin == 0)
                                specialSnowflakeSonicMimic(p)
                            else
                                soundMimic(p)
                            end
                        end
                    end
                
                    --Game modes, yay!
                    if not (p.exiting)

                        --Respawn mode, for those that either wish to play well or commit sudoku
                        if (cv_respawnmimic.value)
                            respawnMimic(p)
                        end

                        --Laps mode, for those that might want it
                        if (cv_lapmimic.value)
                            lapMimic(p)
                        end

                        --Manual mode, maybe I can make this better?
                        if (cv_manualmimic.value)
                            manualMimic(p)
                        end

                        --Random mode, for true insanity!
                        if (cv_randommimic.value)
                            randomMimic(p)
                        end
                    end
                end
            end
		end
	end
end)


--pretty shit

--Manual mimic hud bit. Shown when manual mode is enabled
local function drawManualMimicHud(v, p, x, y, flags)
    --Hud bit
    if (leveltime > p.mimic.manualDelay)
        v.draw(x+72, y+5, v.cachePatch("BYELSTAT"), flags)
    else
        v.draw(x+72, y+5, v.cachePatch("YELSTAT"), flags)
    end
    
    if ((leveltime < p.mimic.manualHudError) and ((leveltime % 2) == 0))
        v.draw(x+72, y+5, v.cachePatch("REDSTAT"), flags&~V_30TRANS)
    end
    
    --Instructions
    if ((leveltime < MANUALINSTRUCTIONTIME) and not knowstheirshit)  --har har, but appropriate nonetheless
        v.drawString(x+20, y-8, "When the icon is \x83green\x80, tap\n\x84\custom3 \x80to change mimics", flags, "small")
    end
end

--Main mimic hud. Shows the stats and character icon
local function drawMimicHud(v, p, x, y, flags)
    --draw the stats
	v.drawString(x+5,y+4, "  Speed: " .. p.kartspeed, flags)
	v.drawString(x+5,y+13, "Weight: " .. p.kartweight, flags)
            
    --draw the mimiced character icon
    if (leveltime >= (p.mimic.newIndicator-(TICRATE/2)))           
        --flicker effect for when a change is approaching/has happened
        if (((leveltime > (p.mimic.nextChange-(2*TICRATE))) and (leveltime < p.mimic.nextChange)) or (leveltime < p.mimic.newIndicator+TICRATE))
            if ((leveltime % 2) == 0)
                v.draw(x, y, v.cachePatch(skins[p.mimic.randomSkin].facemmap), flags, v.getColormap(p.mimic.randomSkin, skins[p.mimic.randomSkin].prefcolor))
            end
        else
            v.draw(x, y, v.cachePatch(skins[p.mimic.randomSkin].facemmap), flags, v.getColormap(p.mimic.randomSkin, skins[p.mimic.randomSkin].prefcolor))
        end
    end
    
    --Manual Mode Hudbit
    if (cv_manualmimic.value)
        drawManualMimicHud(v, p, x, y, flags)
    end
end


--for 3 or more players, a unique HUD variation. This has the special manual bit baked in because lazy
local function drawSplitscreenHudSpecial(v, p, c, x, y, flags)           
    --draw the mimiced character icon
    if (leveltime >= (p.mimic.newIndicator-(TICRATE/2)))           
        --flicker effect for when a change is approaching/has happened
        if (((leveltime > (p.mimic.nextChange-(2*TICRATE))) and (leveltime < p.mimic.nextChange)) or (leveltime < p.mimic.newIndicator+TICRATE))
            if ((leveltime % 2) == 0)
                v.draw(x, y, v.cachePatch(skins[p.mimic.randomSkin].facemmap), flags, v.getColormap(p.mimic.randomSkin, skins[p.mimic.randomSkin].prefcolor))
            end
        else
            v.draw(x, y, v.cachePatch(skins[p.mimic.randomSkin].facemmap), flags, v.getColormap(p.mimic.randomSkin, skins[p.mimic.randomSkin].prefcolor))
        end
    end

    --draw the stats
	v.drawString(x+5,y+4, "  Speed: " .. p.kartspeed, flags, "small")
	v.drawString(x+5,y+9, "Weight: " .. p.kartweight, flags, "small")

	--Special manual mimic bit (again, lazy)
	if (cv_manualmimic.value)
	    local mx = (splitscreen == 1) and 39 or (c.pnum == 2 or c.pnum == 4) and -10 or 39
        if (leveltime > p.mimic.manualDelay)
            v.draw(x+mx, y+4, v.cachePatch("TEMER1"), flags)  --Yes, I used emeralds... it works and it's small
        else
            v.draw(x+mx, y+4, v.cachePatch("TEMER7"), flags)
        end

        if ((leveltime < p.mimic.manualHudError) and ((leveltime % 2) == 0))
            v.draw(x+mx, y+4, v.cachePatch("TEMER5"), flags&~V_30TRANS)
        end
    end
end

local function drawSplitscreenHud(v, p, c)
    local sx = HUDX()-2
    local sy = c.pnum == 2 and 182 or 82
    local splitflags = (c.pnum == 2 and V_SNAPTOBOTTOM or 0)|V_SNAPTOLEFT
        
    if (splitscreen > 1)  -- more than 2 players to deal with
        sx = (c.pnum == 2 or c.pnum == 4) and 240 or 43
        sy = c.pnum > 2 and 185 or 85
        splitflags = (c.pnum > 2 and V_SNAPTOBOTTOM or 0)|((c.pnum == 2 or c.pnum == 4) and V_SNAPTORIGHT or V_SNAPTOLEFT)
        
        drawSplitscreenHudSpecial(v, p, c, sx, sy, V_30TRANS|splitflags)
        return
    end
    
    drawSplitscreenHudSpecial(v, p, c, sx, sy, V_30TRANS|splitflags)
end

--It uh... It draws things. Or something. Look I'm just a line of text okay
hud.add(function(v, p, c)   
    if (p.spectator)
        return
    end
    
    if (cv_mimicstathud.value and cv_mimicenabled.value) and (leveltime > INDICATORTIME)
        if (splitscreen)
           drawSplitscreenHud(v, p, c)
           return
        end

        drawMimicHud(v, p, HUDX(), HUDY, HUDFLAGS)         
	end
end, "game")
